#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>
#include <player.h>
#include <dev1000.h>

extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;

#define MAPPERLESS /*+137 instruction words + MAX_BADS*2 data words. */
#ifdef MAPPERLESS
/*
  Note to MAPPERLESS:
  - This option can be used when FLASH programmer is used:
    a FAT filesystem image is written by a flash programmer starting
    from the third erase block, and bad eraseblocks are skipped.
    (Use the "skip bad blocks" method of programming.)
  - This disk can not be written to, and thus USB is also disabled.
  - At startup MapperlessSetup() bad eraseblocks are scanned and
    remembered, and the normal sector read function is replaced.
  - An image creator software is available on separate request
    (or from VLSI web pages).
  - The first erase block is reserved for firmware, the second
    erase block is reserved to be used freely by the programmer.
 */
#define NO_USB /*With mapperless you can't program content through USB*/

#define MAX_BADS 256 /*Maximum number of bad erase blocks supported.*/
s_int16 numOfBads;      /* remembers the number of bad erase blocks */
u_int32 bads[MAX_BADS]; /* remembers the bad erase block start block numbers */
auto u_int16 MapperlessReadDiskSector(register __i0 u_int16 *buffer,
				      register __reg_a u_int32 sector) {
    /*Leave 2 erase blocks reserved for firmware and optional settings. */
    __y u_int32 sect = sector + (ph->eraseBlockSize<<1);
    register int i = 0;
    while (i < numOfBads) { /* Find mapping between logical and physical */
	if (sect < bads[i])
	    break;
	sect += ph->eraseBlockSize;
	i++;
    }    /* NOTE: currently reads without error correction code! */
    return (u_int16)ph->Read(ph, sect, 1, buffer, NULL);
}

void MapperlessSetup(void) {    /* Setup -- Scan bad erase blocks. */
    register u_int16 i;
    numOfBads = 0;
    /* reserve first two erase blocks for firmware and optional settings. */
    for (i=2; i<ph->eraseBlocks; i++) {
	/* read both spares */
	ph->Read(ph, (u_int32)i * ph->eraseBlockSize, 2, NULL, minifatBuffer);
	if ((minifatBuffer[0/2 + 2] & 0xff) != 0xff ||
	    (minifatBuffer[16/2+ 2] & 0xff) != 0xff) {	    /* bad */
	    bads[numOfBads++] = (u_int32)i * ph->eraseBlockSize;
	    if (numOfBads == MAX_BADS)
		break;
	}
    }    /* Replace normal read function -- bypass mapper. */
    SetHookFunction((u_int16)ReadDiskSector, MapperlessReadDiskSector);
}
#endif /*MAPPERLESS*/


#ifdef NO_USB
void MyUserInterfaceIdleHook(void) {
    if (uiTrigger) {
	uiTrigger = 0;
#ifdef NO_USB
	/*This version of key scan does not cancel play when USB is attached.*/
	KeyScanNoUSB(0x1f); /* with feature button */
#else
	KeyScan(); /* Normal key scan. */
#endif
	/* Update LEDs */
	{
	  register u_int16 leds = LED1;
	  if (player.pauseOn > 0 && (uiTime & 0x06) != 0)
	    leds = 0; /* power led flashes when in pause mode */
	  if (player.randomOn)
	    leds |= LED2;
	  PERIP(GPIO1_ODATA) = (PERIP(GPIO1_ODATA) & ~(LED1 | LED2)) | leds;
	}
    }
}
#endif /*NO_USB*/

void main(void) {

#if 1 /*Perform some extra inits if we are started from SPI boot. */
    InitAudio(); /* goto 3.0x..4.0x */
    PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
    PERIP(INT_ENABLEH) = INTF_DAC;
    ph = FsPhNandCreate(0);
#endif

    /* Set the leds after nand-boot! */
    PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

#ifdef MAPPERLESS
    MapperlessSetup(); /* Scan eraseblocks, replace ReadDiskSector() */
#else
    if (!map) {
	map = FsMapTnCreate(ph, 0); /* tiny mapper */
    }
#endif/*MAPPERLESS*/

    /* Set hooks */
#ifdef NO_USB
    SetHookFunction((u_int16)IdleHook, MyUserInterfaceIdleHook);
#endif /*NO_USB*/

    PlayerVolume();
    while (1) {
#ifndef NO_USB
	if (USBIsAttached()) {
	    MassStorage();
	}
#endif
	if (InitFileSystem() == 0) {
	    minifatInfo.supportedSuffixes = supportedFiles;
	    player.totalFiles = OpenFile(0xffffU);
	    if (player.totalFiles == 0)
		goto noFSnorFiles;

	    player.nextStep = 1;
	    player.nextFile = 0;
	    while (1) {
		if (player.randomOn) {
		    player.currentFile =
			Shuffle(player.totalFiles, player.currentFile);
		} else {
		    player.currentFile = player.nextFile;
		}
		if (player.currentFile < 0)
		    player.currentFile += player.totalFiles;
		if (player.currentFile >= player.totalFiles)
		    player.currentFile -= player.totalFiles;

		if (OpenFile(player.currentFile) < 0) {
		    /* By default the next track is the next track to play */
		    player.nextFile = player.currentFile + 1;

		    /* Some inits moved from PlayCurrentFile() to here to allow
		       easier resume implementation (cs.goTo set to value at start). */
		    //player.pauseOn = 0; /* moved even earlier */
		    player.volumeOffset = -12; /* reset default volume offset */
		    player.ffCount = 0;
		    PlayerVolume();
		    cs.cancel = 0;
		    cs.goTo = -1;
		    cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
		    cs.fastForward = 1; /* reset play speed to normal */

		    PlayCurrentFile();
		} else {
		    player.nextFile = 0;
		}
#ifndef NO_USB
		if (USBIsAttached())
		    break;
#endif
	    }
	} else {
	    /* AudioOutputSamples does not go to low-power mode + poweroff */
	noFSnorFiles:
	    LoadCheck(&cs, 32); /* decrease or increase clock */
	    memset(tmpBuf, 0, sizeof(tmpBuf));
	    AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	    /* When no samples fit, calls the user interface
	       -- handle volume control and power-off */
	}
    }
}
