#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>
#include <player.h>
#include <dev1000.h>

extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;

#ifndef NO_USB
/*
  These defines enable a set of patches to allow or speed up parallel
  programming through USB using VS1000 Linux Programmer software
  (the current version is available from
  http://www.vlsi.fi/support_data/programmer08.zip).
 */
/*+22 words (+13 data words) -- use custom USB serial number */
#define USB_SERIAL_NUMBER
/*+63 words -- faster response when multiple devices in the same bus */
#define PATCH_SCSI_STATUS
/*+234 words -- 'eject' will turn off the unit, checksumming option */
#define PATCH_SCSI_COMMAND
/*-145 words -- leave out checksumming option if you need to save code space */
//#define NO_CHECKSUMMING


#ifdef USB_SERIAL_NUMBER
#include <usblowlib.h>
void MyInitUSBDescriptors(u_int16 initDescriptors) {
#define D_SIZE 26
    static u_int16 myStringDescriptor3[] = {
	(D_SIZE<<8)|03, //type=string
	('2'<<8)|0,	('0'<<8)|0,	('0'<<8)|0,	('0'<<8)|0,
	('1'<<8)|0,	('0'<<8)|0,	('0'<<8)|0,	('0'<<8)|0,
	('2'<<8)|0,	('0'<<8)|0,	('0'<<8)|0,	('0'<<8)|0
    };
#undef D_SIZE
    /* Other descriptors can also be changed. */
    RealInitUSBDescriptors(initDescriptors);
    USB.descriptorTable[3] = myStringDescriptor3;
}
#endif/*USB_SERIAL_NUMBER*/


#ifdef PATCH_SCSI_STATUS
#include <usblowlib.h>
#include <scsi.h>
void MyUSBHandler(void) {
    RealUSBHandler();
    /*Patch for programming multiple devices in the same USB bus */
    if (ScsiState() == SCSI_READY_FOR_COMMAND
	&& (PERIP(USB_CONFIG) & 0x7f)) {
	static int nakked = 0;
	register u_int16 stat = PERIP(USB_EP_ST0 + MSC_BULK_IN_ENDPOINT);
	register u_int16 send = PERIP(USB_EP_SEND0 + MSC_BULK_IN_ENDPOINT);
	if ( (stat & USB_STF_IN_EMPTY) ) {
	    if ((stat & USB_STF_IN_NACK_SENT) &&
		send == ((MSC_BULK_IN_ENDPOINT<<10) | 13)) {
		if (++nakked > 10) {
		    /*Resend with the other DATA toggle */
		    PERIP(USB_EP_SEND0+ MSC_BULK_IN_ENDPOINT) = send | 0x8000U;
		    goto clrn;
		}
		goto clr;
	    }
	} else {
	clrn:
	    nakked = 0;
	clr:
	    PERIP(USB_EP_ST0 + MSC_BULK_IN_ENDPOINT) =
		stat & ~USB_STF_IN_NACK_SENT;
	}
    }
}
#endif/*PATCH_SCSI_STATUS*/

#ifdef PATCH_SCSI_COMMAND
#include <usblowlib.h>
#include <scsi.h>
int startUnit = 1; /*Support eject: scsi start/stop unit.*/
u_int32 ScsiLbab(register __i0 u_int16 *res__lbab3);
auto u_int16 NewDiskProtocolCommand(register __i2 u_int16 *cmd) {
    static u_int16 checksumSave[5];

    /*Note: command length is in hi byte of cmd[0],
      so bytes of the SCSI commands are:
              hi  lo
      word 0  len scsi byte 0 (command)
      word 1  1   2
      word 2  3   4
      word 3  5   6
      word 4  7   8
      READ_10 LBA: 2-5
    */
    __y int c = (cmd[OPERATION_CODE] & 0xff);
    if (c == SCSI_START_STOP_UNIT) {
	startUnit = cmd[2]; /*both eject and start unit bits*/
#ifndef NO_CHECKSUMMING
	/* Support Checksumming */
	{
	    u_int32 t = (((u_int32)cmd[1] << 16) | cmd[2]);
	    if ((t & 0xfc) == 0x5c) {
		memset(checksumSave, 0, sizeof(checksumSave));

		/* Note: InitFileSystem() uses minifatBuffer */
		if (InitFileSystem() == 0) {
		    if (!MscSendCsw(SCSI.Status)) {
			SCSI.State = SCSI_READY_FOR_COMMAND;
		    } else {
			SCSI.State = SCSI_SEND_STATUS;
		    }
		    minifatInfo.longFileName[0] = 0;
		    checksumSave[0] = t >> 8;
		    if (OpenFile(checksumSave[0]) < 0) {
			u_int32 accu = 0;
			register u_int32 p = 0;

			while (p < minifatInfo.fileSize) {
			    register int l = 512;
			    ReadDiskSector(minifatBuffer, FatFindSector(p));
			    if (minifatInfo.fileSize < p + 512) {
				l = minifatInfo.fileSize - p;
				if (l & 1) {
				    minifatBuffer[l/2] &= 0xff00;
				    l++;
				}
			    }
			    accu = CheckSumBlock(accu, minifatBuffer, l / 2);
			    *((u_int32 *)&checksumSave[3]) = accu;
			    p += 512;
			}
			*((u_int32 *)&checksumSave[1]) = minifatInfo.fileSize;
		    }
		    return (u_int16)-1; /* already handled */
		}
	    }
	}
#endif /*!NO_CHECKSUMMING*/
    } else {

	if ((startUnit & 3) == 2) { /* previous command was EJECT MEDIA */
	    /* The next command after STOP UNIT + EJECT MEDIA will not be
	       serviced fully because we will turn ourselves off. */

	    /* Take us 'out of the bus' */
	    PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA;
	    PERIP(USB_CONFIG) = 0x8000U; /* Reset USB */
	    map->Flush(map, 2); /* Flush content */
	    PowerOff();
	}

	if (c == SCSI_REQUEST_SENSE) {
#ifndef NO_CHECKSUMMING	    /* Support Checksumming */
	    /* Read back checksum result in REQUEST_SENSE */
	    if (SCSI.DataTransferLength == 63) {
		memcpy(SCSI.DataBuffer, checksumSave, sizeof(checksumSave));
		memcpyYX(SCSI.DataBuffer+5, minifatInfo.fileName, 6);
		memcpyYX(SCSI.DataBuffer+11, minifatInfo.longFileName,
			 sizeof(SCSI.DataBuffer)-11 /*21, 26 max needed..*/
			 /*sizeof(minifatInfo.longFileName)*/);
		SCSI.DataOutSize = 63;//512;
		SCSI.DataOutBuffer = SCSI.DataBuffer;
		SCSI.State = SCSI_DATA_TO_HOST;   
		return (u_int16)-1;
	    }
#endif /*!NO_CHECKSUMMING*/
	    /* patch a bug -- sense was written to the wrong buffer */
	    SCSI.DataBuffer[0] = 0x7000;
	}
    }
    return cmd[OPERATION_CODE]; /* perform the command */
}
#endif/*PATCH_SCSI_COMMAND*/

#endif/*!NO_USB*/



void main(void) {

#if 1 /*Perform some extra inits if we are started from SPI boot. */
    InitAudio(); /* goto 3.0x..4.0x */
    PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
    PERIP(INT_ENABLEH) = INTF_DAC;
    ph = FsPhNandCreate(0);
#endif

    /* Set the leds after nand-boot! */
    PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

    if (!map) {
	map = FsMapTnCreate(ph, 0); /* tiny mapper */
    }

    PlayerVolume();
    while (1) {
#ifndef NO_USB
	if (USBIsAttached()) {
#ifdef PATCH_SCSI_STATUS
	    SetHookFunction((u_int16)USBHandler, MyUSBHandler);
#endif
#ifdef PATCH_SCSI_COMMAND
	    SetHookFunction((u_int16)MSCPacketFromPC,PatchDiskProtocolCommand);
#endif
#ifdef USB_SERIAL_NUMBER
	    SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);
#endif
	    MassStorage();
	}
#endif
	if (InitFileSystem() == 0) {
	    minifatInfo.supportedSuffixes = supportedFiles;
	    player.totalFiles = OpenFile(0xffffU);
	    if (player.totalFiles == 0)
		goto noFSnorFiles;

	    player.nextStep = 1;
	    player.nextFile = 0;
	    while (1) {
		if (player.randomOn) {
		    player.currentFile =
			Shuffle(player.totalFiles, player.currentFile);
		} else {
		    player.currentFile = player.nextFile;
		}
		if (player.currentFile < 0)
		    player.currentFile += player.totalFiles;
		if (player.currentFile >= player.totalFiles)
		    player.currentFile -= player.totalFiles;

		if (OpenFile(player.currentFile) < 0) {
		    /* By default the next track is the next track to play */
		    player.nextFile = player.currentFile + 1;

		    /* Some inits moved from PlayCurrentFile() to here to allow
		       easier resume implementation. */
		    player.volumeOffset = -12; /* reset default volume offset*/
		    player.ffCount = 0;
		    PlayerVolume();
		    cs.cancel = 0;
		    cs.goTo = -1;
		    cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
		    cs.fastForward = 1; /* reset play speed to normal */

		    PlayCurrentFile();
		} else {
		    player.nextFile = 0;
		}
#ifndef NO_USB
		if (USBIsAttached())
		    break;
#endif
	    }
	} else {
	    /* AudioOutputSamples does not go to low-power mode + poweroff */
	noFSnorFiles:
	    LoadCheck(&cs, 32); /* decrease or increase clock */
	    memset(tmpBuf, 0, sizeof(tmpBuf));
	    AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	    /* When no samples fit, calls the user interface
	       -- handle volume control and power-off */
	}
    }
}
